  /* SPHERES:                          ᯥ⨢
				     Spheres in perspective.
  */
  #include <stdlib.h>
  #include <math.h>
  #include "grasptc.h"

  #define TABLELENGTH 3000
  #define EAST 1
  #define WEST 2
  #define NORTH 3
  #define SOUTH 4
  #define UP 5
  #define DOWN 6

  int m;
  float Xmin=1000, Xmax=-1000, Ymin=1000, Ymax=-1000, C, F;
  struct sph {float X, Y, ze, R; int rdepth;} table[TABLELENGTH];

  float v11, v12, v13, v21, v22, v23, v32, v33, v43,
        f, XC, YC, xC, yC;

  void coeff(float rho, float theta, float phi)
  { float costh, sinth, cosph, sinph;
    costh=cos(theta); sinth=sin(theta);
    cosph=cos(phi); sinph=sin(phi);
    v11=-sinth; v12=-cosph*costh; v13=-sinph*costh;
    v21=costh;  v22=-cosph*sinth; v23=-sinph*sinth;
                v32=sinph;        v33=-cosph;
                                  v43=rho;
  }

  void perspective(float x, float y, float z,
                   float *pX, float *pY, float *pze)
  { float xe, ye, ze;

		     /* न  */ /* Eye coordinates */
    xe = v11*x + v21*y;
    ye = v12*x + v22*y + v32*z;
    ze = v13*x + v23*y + v33*z + v43;

	     /* ࠭ न */  /* Screen coordinates  */
    *pX = xe/ze;
    *pY = ye/ze;
    *pze = ze;
  }

  void sphere1(float x, float y, float z, float r, int n)
  { float X, Y, Xtop, Ytop, R, ze, zetop;
    perspective(x, y, z, &X, &Y, &ze);
    if (X < Xmin) Xmin = X;
    if (X > Xmax) Xmax = X;
    if (Y < Ymin) Ymin = Y;
    if (Y > Ymax) Ymax = Y;
    perspective(x, y, z+r, &Xtop, &Ytop, &zetop);
    R = Ytop - Y;
    table[m].X = X; table[m].Y = Y; table[m].ze = ze; table[m].R = R;
    table[m].rdepth = n;
    if (++m == TABLELENGTH)
    { printf /* "Too many spheres" */
	   ("誮  "); exit(1);
    }
    if (m % 20 == 0) printf("m = %d\n", m);
  }

  void spheres(float x, float y, float z, float r, float n, float dir)
  { int n1=n-1;
    float fr=F*r, cr=C*r;
    if (n > 0)
    { sphere1(x, y, z, r, n);
      if (dir != WEST) spheres(x+cr, y, z, fr, n1, EAST);
      if (dir != EAST) spheres(x-cr, y, z, fr, n1, WEST);
      if (dir != SOUTH) spheres(x, y+cr, z, fr, n1, NORTH);
      if (dir != NORTH) spheres(x, y-cr, z, fr, n1, SOUTH);
      if (dir != DOWN) spheres(x, y, z+cr, fr, n1, UP);
      if (dir != UP) spheres(x, y, z-cr, fr, n1, DOWN);
    }
  }

  float xreal(float X)
  { return xC + f*(X - XC);
  }

  float yreal(float Y)
  { return yC + f*(Y - YC);
  }

  void quicksort(int l, int r)
  /*   ஢ ᠭ  楫 拉 ,  ⮬
     ᫥    "ணࠬ     몥 ".
     This sorting algorithm is discussed in many books,
     such as, for example, PROGRAMS AND DATA STRUCTURES IN C.
  */
  { int i=l, j=r;
    float pivot=table[(l+r)/2].ze;
    struct sph w;
    do
    { while (table[i].ze < pivot) i++;
      while (table[j].ze > pivot) j--;
      if (i > j) break;
      w = table[i]; table[i] = table[j]; table[j] = w;
    } while (++i <= --j);
    if (l < j) quicksort(l, j);
    if (i < r) quicksort(i, r);
  }

  void main(void)
  { float rho, theta, phi, fx, fy, xx, yy, rr;
    int recdepth, i, X, Y, RX, RY, code, fill;
    printf /* "Enter the reduction factors c and f (e.g. 2  0.4): " */
       ("\n 樥 㬥襭 c  f (ਬ, 2 0.4) :  ");
    scanf("%f %f", &C, &F);
    printf /* "Viewing distance (e.g. 12):" */
       ("\nﭨ  (ਬ, 12) :  "); scanf("%f", &rho);
    printf   /* "Give two angles, in degrees." */
	("\n  㣫,  ࠤ.\n");
    printf /* "Theta, measured horizontally from the x-axis (e.g. 20):" */
	("\nTheta -   x  ਧ⠫쭮 ᪮ (ਬ, 20) :  ");
    scanf("%f", &theta); theta *= PI/180;
    printf /* "Phi, measured vertically from the z-axis (e.g. 75): " */
	("\nPhi - 塞   z  ⨪쭮 ᪮"
					      " (ਬ, 75) :  ");
    scanf("%f", &phi); phi *= PI/180;
    printf /* "Recursion depth (e.g. 3): " */
	("\n㡨 ४ᨨ (ਬ, 3) :  "); scanf("%d", &recdepth);
    coeff(rho, theta, phi);
    spheres(0.0, 0.0, 0.0, 1.0, recdepth, 0);
    quicksort(0, m-1);
    initgr();
    xC = x_max/2;
    yC = y_max/2;
    XC = (Xmin+Xmax)/2;
    YC = (Ymin+Ymax)/2;
    fx = x_max/(Xmax-Xmin);
    fy = y_max/(Ymax-Ymin);
    f = 0.9 * (fx < fy ? fx : fy);
    for (i=m-1; i>=0; i--)
    { xx = xreal(table[i].X); X = IX(xx);
      yy = yreal(table[i].Y); Y = IY(yy);
      rr = f * table[i].R; 
      RX = XPIX(rr); RY = YPIX(rr);
      code = table[i].rdepth % 3;
      fill = (code == 2 ? SOLID_FILL :
	         code == 1 ? CLOSE_DOT_FILL : EMPTY_FILL);
      setfillstyle(fill, foregrcolor);
      fillellipse(X, Y, RX, RY);
    }
    endgr();
  }
